@@ -454,7 +454,8 @@ class JsonAccessor(Generic[TSeriesJson]):
         Since the operation might destroy the order of the elements,
         a series containing the offset is also returned.
 
-        :returns: Tuple with SeriesJson (element from the array) and SeriesInt64 (offset of the element)
+        :returns: Tuple with SeriesJson (element from the array) and SeriesInt64 (0-based offset of the
+            element)
 
         .. note::
             Both returned series objects will share same base node, but this node is not
@@ -787,7 +788,7 @@ class JsonAthenaAccessorImpl(Generic[TSeriesJson]):
         stop_expression = self._get_slice_partial_expr(value=key.stop, is_start=False)
 
         values_expression = Expression.construct(
-            "slice(cast({} as array(json)), {}, {})",
+            "try(slice(cast({} as array(json)), {}, {}))",
             self._series_object,
             start_expression,  # startpoint of slice
             Expression.construct('({} - {})', stop_expression, start_expression),  # length of slice
@@ -864,56 +865,57 @@ class JsonAthenaAccessorImpl(Generic[TSeriesJson]):
         if not all(isinstance(value, str) for value in filtering_slice.values()):
             raise TypeError(f'Values in the key dict, should all be strings')
 
-        array_expr = Expression.construct('cast({} as array(json))', self._series_object)
         array_map_expr = Expression.construct(
             'try(cast({} as array(map(varchar, json))))',
             self._series_object
         )
-        filter_condition_exprs = [
+        key_to_extract = [
             Expression.construct(
-                "x[{}] = cast({} as json)",
+                '({}, __x[{}])',
+                Expression.string_value(key),
                 Expression.string_value(key),
-                Expression.string_value(value)
             )
             for key, value in filtering_slice.items()
         ]
-        # filter_array_map_expr: array of maps, but only with the maps that match the filter criteria
-        filter_array_map_expr = Expression.construct(
-            'filter({}, x -> ({}))',
-            array_map_expr,
-            join_expressions(filter_condition_exprs, ' and ')
-        )
-        # filter_array_expr: array with all json objects that match the filter criteria, in the same order
-        # as they appear in the original json array
-        filter_array_expr = Expression.construct('cast({} as array(json))', filter_array_map_expr)
+        key_values_to_match = [
+            Expression.construct(
+                '({}, {})',
+                Expression.string_value(key),
+                Expression.string_value(value),
+            )
+            for key, value in filtering_slice.items()
+        ]
 
+        # extracts only the keys to match
+        transform_array = Expression.construct(
+            'transform({}, __x -> cast(map_from_entries(array[{}]) as json))',
+            array_map_expr,
+            join_expressions(key_to_extract),
+        )
+        element_to_match = Expression.construct(
+            'cast(map_from_entries(array[{}]) as json)',
+            join_expressions(key_values_to_match)
+        )
         if is_start:
             # Find the position in the original array, of the first object that matches the filter criteria
             first_element_pos_expr = Expression.construct(
-                'array_position({}, element_at({}, 1))',
-                array_expr,
-                filter_array_expr
+                'array_position({}, {})',
+                transform_array,
+                element_to_match,
             )
             return first_element_pos_expr
-        # elif not is_start: Find the position in the reversed original array, of the last element that
+
+        # elif not is_start: Find the position in the reversed transformed array, of the last element that
         # matches the filter criteria. This is the last element that matches the criteria in the original
         # array.
-        element_reverse_pos_expr = Expression.construct(
-            'array_position(reverse({}), element_at({}, -1))',
-            array_expr,
-            filter_array_expr
-        )
-        # We need to translate the location in the reverse array to the position in the actual array.
-        # Formula for this: `(array_length + 1 ) - element_reverse_pos_expr`
-        # Additionally we need to add 1, because for historical reasons the end of a filtering slice
+        # Additionally, we need to add 1, because for historical reasons the end of a filtering slice
         # includes the last matched element (unlike a regular slice where the end element is not
         # included).
-        last_element_pos_expr = Expression.construct(
-            '({} + 2 ) - {}',
-            self.get_array_length(),
-            element_reverse_pos_expr
+        return Expression.construct(
+            'array_position(reverse({}), {}) + 1',
+            transform_array,
+            element_to_match,
         )
-        return last_element_pos_expr
 
     def get_array_item(self, key: int) -> 'TSeriesJson':
         """
