@@ -22,6 +22,7 @@ class SeriesUuid(Series):
     **Database support and types**
 
     * Postgres: utilizes the native 'uuid' database type.
+    * Athena: utilizes the generic 'varchar' database type.
     * BigQuery: utilizes the generic 'STRING' database type.
     """
     dtype = 'uuid'
@@ -36,6 +37,7 @@ class SeriesUuid(Series):
     }
 
     supported_value_types = (UUID, str)
+    supported_source_dtypes = ('string',)
 
     @classmethod
     def get_db_dtype(cls, dialect: Dialect) -> Optional[str]:
@@ -70,7 +72,7 @@ class SeriesUuid(Series):
     def dtype_to_expression(cls, dialect: Dialect, source_dtype: str, expression: Expression) -> Expression:
         if source_dtype == 'uuid':
             return expression
-        if source_dtype == 'string':
+        if source_dtype in cls.supported_source_dtypes:
             if is_postgres(dialect):
                 # If the format is wrong, then this will give an error later on, but there is not much we can
                 # do about that here.
